# NOTE: The functions in this file are not tested in chalkpy
# The integration tests in engine/ validate the behavior of load_dataset
from __future__ import annotations

import base64
import json
import os
import uuid
from collections import OrderedDict
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Union, cast
from urllib.parse import urlparse

import pandas as pd
import requests as requests
from typing_extensions import assert_never

from chalk.client import ChalkClient, Dataset, DatasetRevision
from chalk.client.models import (
    ChalkError,
    ColumnMetadata,
    DatasetFilter,
    DatasetResponse,
    OfflineQueryContext,
    QueryStatus,
)
from chalk.features import DataFrame, Feature, FeatureWrapper, deserialize_dtype, ensure_feature
from chalk.features._encoding.pyarrow import pyarrow_to_polars
from chalk.features.feature_set import FeatureSetBase
from chalk.features.pseudofeatures import CHALK_TS_FEATURE, ID_FEATURE, OBSERVED_AT_FEATURE, PSEUDONAMESPACE
from chalk.utils.collections import get_unique_item
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl

    from chalk.client.client_impl import ChalkAPIClientImpl

_DEFAULT_EXECUTOR = ThreadPoolExecutor(16)
_logger = get_logger(__name__)


class ColNameDecoder:
    def decode_col_name(self, col_name: str) -> str:
        if col_name.startswith("__") and col_name.endswith("__"):
            return col_name
        x_split = col_name.split("_")
        if x_split[0] == "ca":
            return "_".join(x_split[1:])
        elif x_split[0] == "cb":
            root_fqn_b32 = x_split[1]
            return base64.b32decode(root_fqn_b32.replace("0", "=").upper()).decode("utf8")
        elif x_split[0] == "cc":
            # Need to implement serialization / deserialization of the state dict
            raise NotImplementedError("Decoding stateful column names are not yet supported")
        else:
            raise ValueError(f"Unexpected identifier: {x_split[0]}")


class DatasetVersion(IntEnum):
    """Format of the parquet file. Used when loading a dataset so that we know what format it is in"""

    # This is the format that bigquery dumps to when specifying an output bucket and output format
    # as part of an (async) query job
    # The output contains extra columns, and all column names are b32 encoded, because
    # bigquery does not support '.' in column names.
    # The client will have to decode column names before loading this data
    # All data, except for feature times, are json encoded
    BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES = 1

    # This is the format returned by the dataset writer in engine/
    DATASET_WRITER = 2

    # This format uses separate columns for the observed at and timestamp columns
    # The observed at column is the actual timestamp from when the observation was observed,
    # whereas the timestamp column is the original timestamp that the user requested
    BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2 = 3

    COMPUTE_RESOLVER_OUTPUT_V1 = 4

    NATIVE_DTYPES = 5
    """This format has feature values decoded with their native data types.
    It does not require json decoding client-side"""


def _parallel_download(uris: List[str], executor: ThreadPoolExecutor, lazy: bool) -> Union[pl.DataFrame, pl.LazyFrame]:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")

    df_futures: list[Future[pl.LazyFrame] | Future[pl.DataFrame]] = []
    for uri in uris:
        if lazy:
            df_futures.append(executor.submit(pl.scan_parquet, uri))
        else:
            df_futures.append(executor.submit(pl.read_parquet, uri))

    dfs = [df.result() for df in df_futures]
    dfs = [x.select(sorted(x.columns)) for x in dfs]
    # Cast the list to be homogenous
    dfs = cast(Union[List[pl.DataFrame], List[pl.LazyFrame]], dfs)
    df = pl.concat(dfs)
    return df


def _load_dataset_from_chalk_writer(uris: List[str]) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    # V1 datasets should contain just a single URI
    # This URI can be read directly
    # We need to filter the features to remove any pseudofeatures
    if len(uris) != 1:
        raise ValueError("v1 datasets should have just a single URI")
    df = pl.read_parquet(uris[0])
    return df


def _decode_column_names(
    column_names: List[str],
) -> Mapping[str, str]:
    ans: Dict[str, str] = {}
    col_name_decoder = ColNameDecoder()
    for x in column_names:
        if x.startswith("__"):
            if x in ("__id__", ID_FEATURE.fqn):
                ans[x] = ID_FEATURE.fqn
            elif x in ("__ts__", CHALK_TS_FEATURE.fqn):
                # Preserve these columns as-is to help with loading the timestamp
                ans[x] = CHALK_TS_FEATURE.fqn
            elif x in ("__observed_at__", "__oat__", OBSERVED_AT_FEATURE.fqn):
                # Preserve these columns as-is to help with loading the timestamp
                ans[x] = OBSERVED_AT_FEATURE.fqn
            # Drop all the other metadata columns
            continue
        feature_name = col_name_decoder.decode_col_name(x)
        if any(feature_name.endswith(f".__{x}__") for x in ("oat", "rat", "observed_at", "replaced_observed_at")):
            # Drop the timestamp metadata from individual features
            continue
        ans[x] = feature_name
    return ans


def _json_decode(x: Optional[str]):
    if x is None:
        return None
    return json.loads(x)


def _load_dataset_bigquery(
    uris: List[str],
    executor: Optional[ThreadPoolExecutor],
    output_feature_fqns: Optional[Sequence[str]],
    output_ts: bool,
    output_id: bool,
    version: DatasetVersion,
    lazy: bool,
    columns: Optional[Sequence[ColumnMetadata]],
) -> Union[pl.DataFrame, pl.LazyFrame]:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    del pl  # unused
    # V2 datasets are in multiple files, and have column names encoded
    # due to DB limitations (e.g. bigquery does not support '.' in column names)
    # In addition, the datasets may contain extra columns (e.g. replaced observed at)
    # All values are JSON encoded
    if executor is None:
        executor = _DEFAULT_EXECUTOR
    df = _parallel_download(uris, executor, lazy)
    return _extract_df_columns(df, output_feature_fqns, output_ts, output_id, version, columns)


def _extract_df_columns(
    df: Union[pl.DataFrame, pl.LazyFrame],
    output_feature_fqns: Optional[Sequence[str]],
    output_ts: bool,
    output_id: bool,
    version: DatasetVersion,
    column_metadata: Optional[Sequence[ColumnMetadata]] = None,
) -> Union[pl.DataFrame, pl.LazyFrame]:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    if version in (
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2,
        DatasetVersion.NATIVE_DTYPES,
    ):
        decoded_col_names = _decode_column_names(df.columns)

        # Select only the columns in decoded_col_names
        df = df.select(list(decoded_col_names.keys()))
        df = df.rename(dict(decoded_col_names))

        # Using an OrderedDict so the order will match the order the user set in the
        # output argument
        expected_cols: Dict[str, pl.Expr] = OrderedDict()
        id_col = pl.col(str(ID_FEATURE))
        if output_id:
            # All dataframes have an __id__ column
            expected_cols[str(ID_FEATURE)] = id_col.alias(str(ID_FEATURE))

        ts_col = pl.col(str(OBSERVED_AT_FEATURE)).fill_null(pl.col(str(CHALK_TS_FEATURE))).dt.with_time_zone("UTC")
        if output_ts:
            # For the ts feature, we want `__oat__` if not null else `__ts__`
            # it is not null; otherwise, we want
            expected_cols[str(CHALK_TS_FEATURE)] = ts_col.alias(str(CHALK_TS_FEATURE))

        if output_feature_fqns is None:
            # If not provided, return all columns, except for the OBSERVED_AT_FEATURE
            # (the REPLACED_OBSERVED_AT was already dropped in _decode_col_names)
            for x in df.columns:
                if x not in expected_cols and not x.startswith(f"{PSEUDONAMESPACE}.") and "chalk_observed_at" not in x:
                    expected_cols[x] = pl.col(x)

        else:
            # Make a best-effort attempt to determine the pkey and ts column fqn from the root namespace
            # of the other features
            root_ns = get_unique_item(
                [x.split(".")[0] for x in df.columns if not x.startswith(f"{PSEUDONAMESPACE}.")], "root_ns"
            )
            ts_feature = None
            pkey_feature = None
            features_cls = None
            if root_ns in FeatureSetBase.registry:
                features_cls = FeatureSetBase.registry[root_ns]
                ts_feature = features_cls.__chalk_ts__
                pkey_feature = features_cls.__chalk_primary__
            for x in output_feature_fqns:
                if features_cls is not None and x in [f.fqn for f in features_cls.features if f.is_has_one]:
                    for col in df.columns:
                        if col.startswith(f"{x}.") and not col.startswith("__"):
                            expected_cols[col] = pl.col(col)
                    continue
                if x == root_ns:
                    for col in df.columns:
                        if col.startswith(root_ns) and not col.startswith("__"):
                            expected_cols[col] = pl.col(col)
                    continue
                if x in expected_cols:
                    continue
                if x in df.columns:
                    if x == str(CHALK_TS_FEATURE):
                        expected_cols[x] = ts_col.alias(x)
                    else:
                        expected_cols[x] = pl.col(x)
                    continue
                if x == str(CHALK_TS_FEATURE) or (ts_feature is not None and x == str(ts_feature)):
                    # The ts feature wasn't returned as the ts feature, but we are able to figure it out from the graph
                    # Alias the ts_col as the ts fqn (or CHALK_TS_FEATURE fqn if that's what was passed in)
                    expected_cols[x] = ts_col.alias(x)
                    continue
                if pkey_feature is not None and x == str(pkey_feature):
                    expected_cols[x] = id_col.alias(x)
                    continue
                else:
                    # We should _never_ hit this as the query should have failed before results are returned
                    # if an invalid feature was requested
                    raise ValueError(f"Feature '{x}' was not found in the results.")

        df = df.select(list(expected_cols.values()))

    elif version == DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1:
        unique_features = set(df.select(pl.col("feature_name").unique())["feature_name"].to_list())
        cols = [
            pl.col("value").filter(pl.col("feature_name") == fqn).first().alias(cast(str, fqn))
            for fqn in unique_features
        ]

        df = df.groupby("pkey").agg(cols)
        decoded_stmts = []
        for col in df.columns:
            if col == "pkey":
                continue
            else:
                decoded_stmts.append(
                    pl.col(col).apply(_json_decode, return_dtype=Feature.from_root_fqn(col).converter.polars_dtype)
                )
        df = df.select(decoded_stmts)
        # it might be a good idea to remember that we used to rename this __id__ column to the primary key
        # We also need to remove columns like feature.__oat__ and feature.__rat__
        df = df.select([col for col in df.columns if not col.endswith("__")])
        return df.select(sorted(df.columns))
    elif version != DatasetVersion.DATASET_WRITER:
        raise ValueError(f"Unsupported version: {version}")

    decoded_stmts = []
    feature_name_to_metadata = None if column_metadata is None else {x.feature_fqn: x for x in column_metadata}
    for col, dtype in zip(df.columns, df.dtypes):
        if version in (
            DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
            DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2,
        ):
            # The parquet file is all JSON-encoded except for the ts column. That is, the only datetime column is for the timestamp,
            # and all other columns are strings
            if isinstance(dtype, pl.Datetime):
                # Assuming that the only datetime column is for timestamps
                decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
            else:
                decoded_stmts.append(pl.col(col).apply(_json_decode))
        elif version == DatasetVersion.NATIVE_DTYPES:
            if column_metadata is None:
                raise ValueError("The columns must be provided if the dataset type is NATIVE_DTYPES")
            # We already decoded the column names so matching against the fqn
            if col == CHALK_TS_FEATURE or col == OBSERVED_AT_FEATURE:
                decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
            elif col == ID_FEATURE:
                # The pkey is already decoded properly -- it's always an int or str
                decoded_stmts.append(pl.col(col))
            else:
                assert feature_name_to_metadata is not None
                if col not in feature_name_to_metadata:
                    _logger.info(
                        (
                            f"Column '{col}' was included in the dataset but was not returned in the ColumnMetadata. "
                            "This feature might not be casted correctly."
                        )
                    )
                    if isinstance(dtype, pl.Datetime):
                        # It's probably a TS feature
                        decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
                    else:
                        decoded_stmts.append(pl.col(col))
                else:
                    col_metadata = feature_name_to_metadata[col]
                    polars_dtype = pyarrow_to_polars(deserialize_dtype(col_metadata.dtype), col)
                    # Don't attempt to cast list and struct types -- it probably won't work
                    # Instead, we should load the dataset via pyarrow, rather than via polars
                    col_expr = pl.col(col)
                    if dtype != polars_dtype and not isinstance(polars_dtype, (pl.Struct, pl.List)):
                        col_expr = col_expr.cast(polars_dtype, strict=True)
                    decoded_stmts.append(col_expr)
        else:
            raise ValueError(f"Unsupported version: {version}")
    return df.select(decoded_stmts)


def load_dataset(
    uris: List[str],
    version: Union[DatasetVersion, int],
    output_features: Optional[Sequence[Union[str, Feature, FeatureWrapper, Any]]] = None,
    output_id: bool = True,
    output_ts: bool = True,
    executor: Optional[ThreadPoolExecutor] = None,
    lazy: bool = False,
    columns: Optional[Sequence[ColumnMetadata]] = None,
) -> Union[pl.DataFrame, pl.LazyFrame]:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[runtime]")
    del pl  # Unused
    if not isinstance(version, DatasetVersion):
        try:
            version = DatasetVersion(version)
        except ValueError:
            raise ValueError(
                (
                    f"The dataset version ({version}) is not supported by this installed version of the Chalk client. "
                    "Please upgrade your chalk client and try again."
                )
            )
    if version == DatasetVersion.DATASET_WRITER:
        return _load_dataset_from_chalk_writer(uris)
    output_feature_fqns = (
        None
        if output_features is None
        else [x if isinstance(x, str) else ensure_feature(x).root_fqn for x in output_features]
    )
    if version in (
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2,
        DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1,
        DatasetVersion.NATIVE_DTYPES,
    ):
        return _load_dataset_bigquery(
            uris,
            executor,
            version=version,
            output_feature_fqns=output_feature_fqns,
            output_id=output_id,
            output_ts=output_ts,
            lazy=lazy,
            columns=columns,
        )
    assert_never(version)


class DatasetRevisionImpl(DatasetRevision):
    def __init__(
        self,
        revision_id: uuid.UUID,
        creator_id: str,
        outputs: List[str],
        givens_uri: str,
        status: QueryStatus,
        filters: DatasetFilter,
        num_partitions: int,
        output_uris: str,
        output_version: int,
        client: ChalkAPIClientImpl,
        num_bytes: Optional[int] = None,
        created_at: Optional[datetime] = None,
        started_at: Optional[datetime] = None,
        terminated_at: Optional[datetime] = None,
        dataset_name: Optional[str] = None,
        dataset_id: Optional[uuid.UUID] = None,
    ):
        self.revision_id = revision_id
        self.creator_id = creator_id
        self.outputs = outputs
        self.givens_uri = givens_uri
        self.status = status
        self.filters = filters
        self.num_partitions = num_partitions
        self.output_uris = output_uris
        self.output_version = output_version
        self.num_bytes = num_bytes
        self.created_at = created_at
        self.started_at = started_at
        self.terminated_at = terminated_at
        self.dataset_name = dataset_name
        self.dataset_id = dataset_id
        self._client = client

    @property
    def data_as_polars(self) -> pl.LazyFrame:
        return self.get_data_as_polars()

    def get_data_as_polars(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> pl.LazyFrame:
        if context is None:
            context = OfflineQueryContext()
        _logger.info(f"loading polars LazyFrame for DatasetRevision {self.revision_id}")
        return self._client._await_offline_query_job(
            job_id=self.revision_id,
            outputs=self.outputs,
            lazy=True,
            context=context,
            preview_deployment_id=preview_deployment_id,
            output_id=output_id,
            output_ts=output_ts,
        )

    @property
    def data_as_pandas(self) -> pd.DataFrame:
        return self.get_data_as_pandas()

    def get_data_as_pandas(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> pd.DataFrame:
        if context is None:
            context = OfflineQueryContext()
        _logger.info(f"loading pandas DataFrame for DatasetRevision {self.revision_id}")
        return (
            self._client._await_offline_query_job(
                job_id=self.revision_id,
                outputs=self.outputs,
                lazy=True,
                context=context,
                preview_deployment_id=preview_deployment_id,
                output_id=output_id,
                output_ts=output_ts,
            )
            .collect()
            .to_pandas()
        )

    @property
    def data_as_dataframe(self) -> DataFrame:
        return self.get_data_as_dataframe()

    def get_data_as_dataframe(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> DataFrame:
        if context is None:
            context = OfflineQueryContext()
        _logger.info(f"loading Chalk DataFrame for DatasetRevision {self.revision_id}")
        return DataFrame(
            data=self._client._await_offline_query_job(
                job_id=self.revision_id,
                outputs=self.outputs,
                lazy=True,
                context=context,
                preview_deployment_id=preview_deployment_id,
                output_id=output_id,
                output_ts=output_ts,
            )
        )

    def download_data(
        self,
        path: str,
        num_executors: Optional[int] = None,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
    ):
        if context is None:
            context = OfflineQueryContext()
        urls = self._client._await_offline_query_job(
            job_id=self.revision_id,
            outputs=self.outputs,
            lazy=True,
            context=context,
            preview_deployment_id=preview_deployment_id,
            urls_only=True,
        )
        if num_executors:
            executor = ThreadPoolExecutor(16)
        else:
            executor = _DEFAULT_EXECUTOR

        def _download_data(url: str, directory_path: str):
            r = requests.get(url)
            parse = urlparse(url)
            destination_filepath = "/".join(parse.path.split("/")[4:])
            destination_directory = os.path.join(directory_path, os.path.dirname(destination_filepath))
            os.makedirs(destination_directory, exist_ok=True)
            with open(f"{directory_path}/{destination_filepath}", "wb") as f:
                f.write(r.content)

        for url in urls:
            executor.submit(_download_data, url, path).result()

    def __repr__(self) -> str:
        if self.dataset_name:
            return f"DatasetRevision(dataset_name='{self.dataset_name}', revision_id='{self.revision_id}', status='{self.status.value}')"
        return f"DatasetRevision(revision_id='{self.revision_id}')"


class DatasetImpl(Dataset):
    def __init__(
        self,
        is_finished: bool,
        version: int,
        revisions: List[DatasetRevisionImpl],
        client: ChalkClient,
        dataset_id: Optional[uuid.UUID] = None,
        dataset_name: Optional[str] = None,
        errors: Optional[List[ChalkError]] = None,
    ):
        self._is_finished = is_finished
        self._version = version
        self._revisions = revisions
        self._dataset_id = dataset_id
        self._dataset_name = dataset_name
        self._errors = errors
        self._client = client

    @property
    def is_finished(self) -> bool:
        return self._is_finished

    @property
    def version(self) -> int:
        return self._version

    @property
    def revisions(self) -> List[DatasetRevisionImpl]:
        return self._revisions

    @property
    def dataset_id(self) -> Union[uuid.UUID, None]:
        return self._dataset_id

    @property
    def dataset_name(self) -> Union[str, None]:
        return self._dataset_name

    @property
    def errors(self) -> Optional[List[ChalkError]]:
        return self._errors

    @property
    def data_as_polars(self) -> pl.LazyFrame:
        return self.get_data_as_polars()

    def get_data_as_polars(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> pl.LazyFrame:
        if len(self._revisions) == 0:
            raise IndexError("No revisions exist for dataset")
        return self._revisions[-1].get_data_as_polars(
            context=context,
            preview_deployment_id=preview_deployment_id,
            output_id=output_id,
            output_ts=output_ts,
        )

    @property
    def data_as_pandas(self) -> pd.DataFrame:
        return self.get_data_as_pandas()

    def get_data_as_pandas(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> pd.DataFrame:
        if len(self._revisions) == 0:
            raise IndexError("No revisions exist for dataset")
        return self._revisions[-1].get_data_as_pandas(
            context=context,
            preview_deployment_id=preview_deployment_id,
            output_id=output_id,
            output_ts=output_ts,
        )

    @property
    def data_as_dataframe(self) -> DataFrame:
        return self.get_data_as_dataframe()

    def get_data_as_dataframe(
        self,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        output_id: bool = False,
        output_ts: bool = False,
    ) -> DataFrame:
        if len(self._revisions) == 0:
            raise IndexError("No revisions exist for dataset")
        return self._revisions[-1].get_data_as_dataframe(
            context=context,
            preview_deployment_id=preview_deployment_id,
            output_id=output_id,
            output_ts=output_ts,
        )

    def download_data(
        self,
        path: str,
        num_executors: Optional[int] = 16,
        context: Optional[OfflineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
    ):
        return self._revisions[-1].download_data(
            path=path, num_executors=num_executors, context=context, preview_deployment_id=preview_deployment_id
        )

    def __repr__(self) -> str:
        if self._errors:
            return f"Dataset(name='{self._dataset_name}', version='{self._version}', errors='{self._errors}')"
        if self._dataset_name:
            return f"Dataset(name='{self._dataset_name}', version='{self._version}')"
        return f"Dataset()"


def dataset_from_response(response: DatasetResponse, client: ChalkAPIClientImpl) -> DatasetImpl:
    revisions = [
        DatasetRevisionImpl(
            revision_id=revision.revision_id,
            creator_id=revision.creator_id,
            outputs=revision.outputs,
            givens_uri=revision.givens_uri,
            status=revision.status,
            filters=revision.filters,
            num_partitions=revision.num_partitions,
            output_uris=revision.output_uris,
            output_version=revision.output_version,
            num_bytes=revision.num_bytes,
            client=client,
            created_at=revision.created_at,
            started_at=revision.started_at,
            terminated_at=revision.terminated_at,
            dataset_name=revision.dataset_name,
            dataset_id=revision.dataset_id,
        )
        for revision in response.revisions
    ]
    return DatasetImpl(
        is_finished=response.is_finished,
        version=response.version,
        revisions=revisions,
        client=client,
        dataset_id=response.dataset_id,
        dataset_name=response.dataset_name,
        errors=response.errors,
    )



from __future__ import annotations

import functools
import warnings
from typing import TYPE_CHECKING, Any, Generic, Optional, Protocol, Sequence, Type, TypeVar, Union, cast

import pyarrow as pa

from chalk.features._encoding.json import structure_json_to_primitive, unstructure_primitive_to_json
from chalk.features._encoding.missing_value import MissingValueStrategy
from chalk.features._encoding.primitive import TPrimitive
from chalk.features._encoding.pyarrow import pyarrow_to_polars, pyarrow_to_primitive, rich_to_pyarrow
from chalk.features._encoding.rich import structure_primitive_to_rich, unstructure_rich_to_primitive
from chalk.utils.collections import unwrap_annotated_if_needed
from chalk.utils.json import TJSON
from chalk.utils.log_with_context import get_logger

_logger = get_logger(__name__)

if TYPE_CHECKING:
    import polars as pl

_TRich = TypeVar("_TRich")
_TRichCo = TypeVar("_TRichCo", covariant=True)
_TRichCon = TypeVar("_TRichCon", contravariant=True)

_TPrim = TypeVar("_TPrim", bound=TPrimitive)
_TPrimCo = TypeVar("_TPrimCo", bound=TPrimitive, covariant=True)
_TPrimCon = TypeVar("_TPrimCon", bound=TPrimitive, contravariant=True)


class TEncoder(Protocol[_TPrimCo, _TRichCon]):
    def __call__(self, value: _TRichCon, /) -> _TPrimCo:
        ...


class TDecoder(Protocol[_TPrimCon, _TRichCo]):
    def __call__(self, value: _TPrimCon, /) -> _TRichCo:
        ...


class FeatureConverter(Generic[_TPrim, _TRich]):
    def __init__(
        self,
        name: str,
        is_nullable: bool,
        rich_default: Union[_TRich, ellipsis] = ...,
        primitive_default: Union[_TPrim, ellipsis] = ...,
        rich_type: Union[Type[_TRich], ellipsis] = ...,
        pyarrow_dtype: Optional[pa.DataType] = None,
        encoder: Optional[TEncoder[_TPrim, _TRich]] = None,
        decoder: Optional[TDecoder[_TPrim, _TRich]] = None,
    ) -> None:
        self._name = name
        self._rich_type = unwrap_annotated_if_needed(rich_type)
        if pyarrow_dtype is None:
            if rich_type is ...:
                raise ValueError("Either the `rich_type` or `pyarrow_dtype` must be provided")
            pyarrow_dtype = rich_to_pyarrow(rich_type, name)

        if rich_type is ...:
            if rich_default != ...:
                raise ValueError(
                    "The `rich_default` cannot be used without the `rich_type`. Perhaps specify the `primitive_default` instead?"
                )
            if is_nullable and primitive_default is ...:
                primitive_default = cast(_TPrim, None)

        else:
            if primitive_default != ...:
                raise ValueError(
                    "The `primitive_default` cannot be used when specifying the `rich_type`. Instead, specify the `rich_default`."
                )
            if is_nullable and rich_default is ...:
                rich_default = cast(_TRich, None)

        self._pyarrow_dtype = pyarrow_dtype
        if pa.types.is_struct(pyarrow_dtype) and is_nullable:
            # Structs are never nullable. Their members can be set to null, however.
            warnings.warn(
                UserWarning(
                    (
                        f"Feature '{self._name}' is an nullable struct (e.g. Optional[...]). A `null` value for a struct "
                        "is equivalent to `null` values for each member of the struct. To fix this warning, remove the "
                        "`Optional[...]` annotation for this feature, and instead add `Optional[...]` to each struct member "
                    )
                )
            )
            is_nullable = False
        self._is_nullable = is_nullable
        self._primitive_type = pyarrow_to_primitive(self._pyarrow_dtype, name)
        if rich_type is ...:
            if encoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
            if decoder is not None:
                raise ValueError("An encoder cannot be specified without also specifying the `rich_type`")
        self._encoder = encoder
        self._decoder = decoder
        self._rich_default = rich_default
        if primitive_default is ... and rich_default != ...:
            # The missing value strategy doesn't really matter because rich_default is not missing
            primitive_default = self.from_rich_to_primitive(rich_default, missing_value_strategy="allow")
        self._primitive_default = primitive_default

    def _to_primitive(self, val: _TRich) -> _TPrim:
        if val is None or self._encoder is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                x = unstructure_rich_to_primitive(val)
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}'"
                ) from e
            if x is None and not self._is_nullable:
                raise ValueError(f"Feature '{self._name}' is null, but it cannot be nullable")
            try:
                return cast(_TPrim, structure_primitive_to_rich(x, cast(Type[_TRich], self.primitive_type)))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}"
                ) from e
        return self._encoder(val)

    def _from_prim(self, val: Union[_TPrim, _TRich]) -> _TRich:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        if val is None:
            # Structuring null values to the primitive type to ensure that a singular null for an entire struct
            # is propagated to individual struct fields -- e.g.
            # class LatLong:
            #     lat: Optional[float]
            #     long: Optional[float]
            # then self._from_prim(None) == LatLong(None, None)
            # Using self.primitive_type, rather than self._rich_type, as the primitive type
            # might not be registered on the converter for custom classes
            try:
                val = structure_primitive_to_rich(cast(_TPrim, val), cast(Type[_TRich], self.primitive_type))
            except (TypeError, ValueError) as e:
                raise TypeError(
                    f"Could not convert '{val}' to `{self.primitive_type}` for feature '{self._name}'"
                ) from e
        if self._decoder is None:
            try:
                return structure_primitive_to_rich(cast(_TPrim, val), self._rich_type)
            except (TypeError, ValueError) as e:
                raise TypeError(f"Could not convert '{val}' to `{self._rich_type}` for feature '{self._name}'") from e
        if isinstance(val, self._rich_type):
            return cast(_TRich, val)
        return self._decoder(cast(_TPrim, val))

    def from_rich_to_pyarrow(
        self,
        values: Sequence[Union[_TRich, ellipsis, None]],
        /,
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> pa.Array:
        prim_values = [self.from_rich_to_primitive(x, missing_value_strategy) for x in values]
        return self.from_primitive_to_pyarrow(prim_values)

    def from_rich_to_primitive(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> _TPrim:
        # Ensure that the rich value is indeed the rich type
        # For example, if a string is passed in for a datetime value, convert it into a datetime
        if self.is_value_missing(value):
            if missing_value_strategy == "allow":
                warnings.warn(
                    UserWarning(f"Allowing missing value for feature '{self._name}' with strategy 'default_or_allow'")
                )
                return cast(_TPrim, value)
            elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                if self.has_default:
                    return self.primitive_default
                elif missing_value_strategy == "default_or_error":
                    raise TypeError(
                        f"The value for feature '{self._name}' is missing, and this feature has no default value."
                    )
                else:
                    warnings.warn(
                        UserWarning(
                            f"Allowing missing value for feature '{self._name}' with strategy 'default_or_allow'"
                        )
                    )
                    return cast(_TPrim, value)
            elif missing_value_strategy == "error":
                raise TypeError(
                    f"The value for feature '{self._name}' is missing, but `replace_missing_with_defaults` was set to False."
                )
            else:
                raise ValueError(
                    (
                        f"Unsupported missing value strategy: {missing_value_strategy}. "
                        "It must be one of 'allow', 'default_or_allow', 'default_or_error', or 'error'."
                    )
                )
        value = self.from_primitive_to_rich(cast(_TPrim, value))
        return self._to_primitive(value)

    def from_rich_to_json(
        self,
        value: Union[_TRich, ellipsis, None],
        missing_value_strategy: MissingValueStrategy = "default_or_allow",
    ) -> TJSON:
        prim_val = self.from_rich_to_primitive(value, missing_value_strategy)
        return self.from_primitive_to_json(prim_val)

    def from_pyarrow_to_rich(self, values: pa.Array, /) -> Sequence[_TRich]:
        return [self.from_primitive_to_rich(x) for x in values.to_pylist()]

    def from_pyarrow_to_json(self, values: pa.Array) -> Sequence[TJSON]:
        return [self.from_primitive_to_json(x) for x in self.from_pyarrow_to_primitive(values)]

    def from_pyarrow_to_primitive(self, values: pa.Array) -> Sequence[_TPrim]:
        return values.to_pylist()

    def from_primitive_to_rich(self, value: Union[_TPrim, _TRich]) -> _TRich:
        return self._from_prim(value)

    def from_primitive_to_pyarrow(self, value: Sequence[_TPrim]) -> pa.Array:
        x = pa.array([None if x is ... else x for x in value], type=self._pyarrow_dtype)
        return x

    def from_primitive_to_json(self, value: TPrimitive) -> TJSON:
        return unstructure_primitive_to_json(value)

    def from_json_to_rich(self, value: TJSON) -> _TRich:
        prim_val = self.from_json_to_primitive(value)
        return self.from_primitive_to_rich(prim_val)

    def from_json_to_pyarrow(self, values: Sequence[TJSON]) -> pa.Array:
        primitive_vals = [cast(_TPrim, self.from_json_to_primitive(x)) for x in values]
        return self.from_primitive_to_pyarrow(primitive_vals)

    def from_json_to_primitive(self, value: Union[TJSON, TPrimitive]) -> _TPrim:
        try:
            return cast(_TPrim, structure_json_to_primitive(value, self._primitive_type))
        except (ValueError, TypeError) as e:
            raise TypeError(f"Could not convert '{value}' to `{self._primitive_type}`") from e

    @property
    def pyarrow_dtype(self):
        return self._pyarrow_dtype

    @property
    def rich_type(self) -> Type[_TRich]:
        if self._rich_type is ...:
            raise ValueError(
                "Rich types cannot be used as the FeatureConverter was created without providing a `rich_type`"
            )
        return self._rich_type

    @property
    def primitive_type(self) -> Type[TPrimitive]:
        return self._primitive_type

    @functools.cached_property
    def polars_dtype(self) -> pl.PolarsDataType:
        return pyarrow_to_polars(self.pyarrow_dtype, self._name)

    @property
    def has_default(self):
        # primitive default is set <==> rich default is also set or no rich type was provided
        return self._primitive_default != ...

    @property
    def rich_default(self) -> _TRich:
        if self._rich_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return self._rich_default

    @property
    def primitive_default(self) -> _TPrim:
        if self._primitive_default is ...:
            raise ValueError(f"Feature '{self._name}' has no default value")
        return self._primitive_default

    def is_value_missing(self, value: Any):
        """Returns whether the ``value`` should be treated as a "missing" value"""
        if value is ...:
            # Ellipsis is always missing
            return True
        if value is None:
            # All nullable args have a default (``None`` if not otherwise specified)
            if pa.types.is_struct(self.pyarrow_dtype):
                # Nones are not missing for structs
                return False
            return not self._is_nullable
        return False



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from sqlalchemy.engine.url import URL


class RedshiftSourceImpl(BaseSQLSource):
    def __init__(
        self,
        host: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        try:
            import redshift_connector
        except ImportError:
            raise missing_dependency_exception("chalkpy[redshift]")
        del redshift_connector
        self.host = host or load_integration_variable(name="REDSHIFT_HOST", integration_name=name)
        self.db = db or load_integration_variable(name="REDSHIFT_DB", integration_name=name)
        self.user = user or load_integration_variable(name="REDSHIFT_USER", integration_name=name)
        self.password = password or load_integration_variable(name="REDSHIFT_PASSWORD", integration_name=name)
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        from sqlalchemy.engine.url import URL

        return URL.create(
            drivername="redshift+psycopg2",
            username=self.user,
            password=self.password,
            host=self.host,
            database=self.db,
        )



import ast
import inspect
import textwrap
from typing import Any, Callable, Dict, Optional

from chalk.df.ast_parser import convert_slice, eval_converted_expr


class ResolverAnnotationParser:
    def __init__(self, resolver: Callable, glbs: Optional[Dict[str, Any]], lcls: Optional[Dict[str, Any]]):
        self.resolver = resolver
        self.glbs = glbs
        self.lcls = lcls

        self._args = {arg.arg: arg for arg in self._get_resolver_args()}

    def _get_resolver_args(self):
        source = inspect.getsource(self.resolver)
        parsed_source = ast.parse(textwrap.dedent(source))
        assert len(parsed_source.body) == 1
        function_def = parsed_source.body[0]
        assert isinstance(function_def, ast.FunctionDef)
        args = function_def.args
        all_args = [*args.posonlyargs, *args.args, *args.kwonlyargs]
        return all_args

    def parse_annotation(self, name: str):
        arg = self._args[name]
        annotation = arg.annotation
        assert (
            annotation is not None
        ), f"Argument '{name}' for resolver '{self.resolver.__name__}' was not defined with a type annotation."
        if isinstance(annotation, ast.Constant):
            val = annotation.value
            assert isinstance(val, str), "if a literal annotation, it must be a string"
            # string of type annotation
            val = ast.parse(val, mode="eval")
            assert isinstance(val, ast.Module)
            annotation = val.body
        if isinstance(annotation, ast.Subscript):
            # All fancy ast parsing would appear within the subscript of a df __getitem__
            annotation = ast.Subscript(
                value=annotation.value,
                slice=convert_slice(annotation.slice),
                ctx=annotation.ctx,
            )
        return eval_converted_expr(annotation, self.glbs, self.lcls)



import dataclasses
import json
import os
import re
import sys
from pathlib import Path
from typing import List, Optional

from chalk._monitoring.Chart import Chart
from chalk._monitoring.gql_conversion import convert_chart
from chalk._version import __version__
from chalk.config.project_config import load_project_config
from chalk.features import FeatureSetBase
from chalk.features.resolver import Resolver, SinkResolver, StreamResolver
from chalk.importer import FailedImport
from chalk.parsed._graph_validation import validate_graph
from chalk.parsed.duplicate_input_gql import (
    ChalkPYInfo,
    EnvironmentSettingsGQL,
    FeatureSettings,
    MetadataSettings,
    ProjectSettingsGQL,
    UpsertGraphGQL,
    ValidationSettings,
)
from chalk.parsed.json_conversions import convert_type_to_gql
from chalk.utils import paths


def _is_relative_to(x: Path, other: Path) -> bool:
    try:
        x.relative_to(other)
        return True
    except ValueError:
        return False


def get_registered_types_as_json(scope_to: Path, failed: List[FailedImport], indent: int = 2) -> str:
    features = [
        convert_type_to_gql(feature)
        for x in FeatureSetBase.registry.values()
        if x.__module__ in sys.modules and _is_relative_to(paths.get_classpath(x), scope_to)
        for feature in x.features
        if not feature.is_autogenerated and not feature.no_display
    ]
    stream_resolvers = [
        convert_type_to_gql(t) for t in StreamResolver.registry if _is_relative_to(Path(t.filename), scope_to)
    ]
    resolvers = [convert_type_to_gql(t) for t in Resolver.registry if _is_relative_to(Path(t.filename), scope_to)]
    sink_resolvers = [
        convert_type_to_gql(t) for t in SinkResolver.registry if _is_relative_to(Path(t.filename), scope_to)
    ]
    charts = [convert_chart(chart) for chart in Chart._registry]

    def read_packages(filename: str) -> Optional[List[str]]:
        reqs = list()
        try:
            with open(filename) as f:
                for r in f.readlines():
                    cleaned = re.sub("#.*", "", r).rstrip("\n").strip()
                    if cleaned != "":
                        reqs.append(cleaned)
            return reqs
        except OSError:
            return None

    config = load_project_config()
    if config is not None:
        config = ProjectSettingsGQL(
            project=config.project,
            environments=config.environments
            and [
                EnvironmentSettingsGQL(
                    id=i,
                    runtime=e.runtime,
                    requirements=e.requirements,
                    dockerfile=e.dockerfile,
                    requiresPackages=None
                    if e.requirements is None
                    else read_packages(
                        os.path.join(
                            os.path.dirname(config.local_path),
                            e.requirements,
                        )
                    ),
                )
                for i, e in config.environments.items()
            ],
            validation=ValidationSettings(
                feature=FeatureSettings(
                    metadata=[
                        MetadataSettings(name=m.name, missing=m.missing) for m in config.validation.feature.metadata
                    ]
                    if config.validation.feature.metadata
                    else None
                )
                if config.validation.feature
                else None
            )
            if config.validation
            else None,
        )

    graph = UpsertGraphGQL(
        streams=stream_resolvers,
        sinks=sink_resolvers,
        features=features,
        config=config,
        failed=failed,
        resolvers=resolvers,
        charts=charts,
        chalkpy=ChalkPYInfo(version=__version__),
    )
    errors = validate_graph(graph)
    graph.errors = errors
    return json.dumps(
        dataclasses.asdict(graph),
        indent=indent,
    )



from __future__ import annotations

import enum
import json
from typing import List, Literal, Optional, Union

import pyarrow as pa
from pydantic import BaseConfig, BaseModel, Field
from typing_extensions import Annotated

__all__ = [
    "deserialize_dtype",
    "serialize_dtype",
    "serialize_pyarrow_dtype",
]


class _DTypeCode(enum.IntEnum):
    # NULL = 1
    BOOL = 2
    INT8 = 3
    INT16 = 4
    INT32 = 5
    INT64 = 6
    UINT8 = 7
    UINT16 = 8
    UINT32 = 9
    UINT64 = 10
    FLOAT16 = 11
    FLOAT32 = 12
    FLOAT64 = 13
    TIME32 = 14  # time unit
    TIME64 = 15  # time unit
    TIMESTAMP = 16  # time unit, time zone
    DATE32 = 17
    DATE64 = 18
    DURATION = 19  # time unit
    BINARY = 20  # length
    STRING = 21
    LARGE_BINARY = 22
    LARGE_STRING = 23
    # Decimal is not yet supported by chalk
    # DECIMAL128 = 24  # precision, scale
    LIST = 25  # value type, size
    LARGE_LIST = 26  # value type
    STRUCT = 27  # fields


class _TimeUnitCode(str, enum.Enum):
    SECOND = "s"
    MILLISECONDS = "ms"
    MICROSECONDS = "us"
    NANOSECONDS = "ns"


class _BaseSerializedDType(BaseModel):
    type_code: _DTypeCode

    def to_pyarrow_dtype(self) -> pa.DataType:
        raise NotImplementedError

    class Config(BaseConfig):
        use_enum_values = True


class _SingletonDType(_BaseSerializedDType):
    type_code: Literal[
        _DTypeCode.BOOL,
        _DTypeCode.INT8,
        _DTypeCode.INT16,
        _DTypeCode.INT32,
        _DTypeCode.INT64,
        _DTypeCode.UINT8,
        _DTypeCode.UINT16,
        _DTypeCode.UINT32,
        _DTypeCode.UINT64,
        _DTypeCode.FLOAT16,
        _DTypeCode.FLOAT32,
        _DTypeCode.FLOAT64,
        _DTypeCode.DATE32,
        _DTypeCode.DATE64,
        _DTypeCode.STRING,
        _DTypeCode.LARGE_STRING,
        _DTypeCode.LARGE_BINARY,
    ]

    def to_pyarrow_dtype(self) -> pa.DataType:
        if self.type_code == _DTypeCode.BOOL:
            return pa.bool_()
        try:
            return getattr(pa, self.type_code.name.lower())()
        except AttributeError:
            raise ValueError(f"Unsupported dtype: {self.type_code}") from None


class _TimeUnitDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.TIME32, _DTypeCode.TIME64, _DTypeCode.DURATION]
    time_unit: _TimeUnitCode

    def to_pyarrow_dtype(self) -> pa.DataType:
        return getattr(pa, self.type_code.name.lower())(self.time_unit)


class _TimestampDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.TIMESTAMP]
    time_unit: _TimeUnitCode
    timezone: Optional[str]

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.timestamp(self.time_unit, self.timezone)


class _BinaryDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.BINARY]
    length: int

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.binary(self.length)


class _ListDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.LIST]
    inner_dtype: _SerializedDType
    length: int

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.list_(self.inner_dtype.to_pyarrow_dtype(), self.length)


class _LargeListDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.LARGE_LIST]
    inner_dtype: _SerializedDType

    def to_pyarrow_dtype(self) -> pa.DataType:
        return pa.large_list(self.inner_dtype.to_pyarrow_dtype())


class _StructField(BaseModel):
    name: str
    dtype: _SerializedDType
    nullable: bool = True

    def to_pyarrow_dtype(self) -> pa.Field:
        return pa.field(self.name, self.dtype.to_pyarrow_dtype(), self.nullable)


class _StructDType(_BaseSerializedDType):
    type_code: Literal[_DTypeCode.STRUCT]
    fields: List[_StructField]

    def to_pyarrow_dtype(self) -> pa.Field:
        return pa.struct([x.to_pyarrow_dtype() for x in self.fields])


_SerializedDType = Annotated[
    Union[_SingletonDType, _TimeUnitDType, _TimestampDType, _BinaryDType, _ListDType, _LargeListDType, _StructDType],
    Field(discriminator="type_code"),
]

_StructField.update_forward_refs()
_ListDType.update_forward_refs()
_LargeListDType.update_forward_refs()


class _Deserializer(BaseModel):
    dtype: _SerializedDType


def deserialize_dtype(serialized_dtype: str) -> pa.DataType:
    deserializer = _Deserializer(dtype=json.loads(serialized_dtype))
    pa_dtype = deserializer.dtype.to_pyarrow_dtype()
    return pa_dtype


def serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
    """Deprecated -- use `serialize_dtype` instead. For backwards compatibility with Engine"""
    return _serialize_pyarrow_dtype(pa_dtype)


def serialize_dtype(pa_dtype: pa.DataType) -> str:
    return _serialize_pyarrow_dtype(pa_dtype).json()


def _serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
    if pa.types.is_boolean(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.BOOL)
    if pa.types.is_uint8(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.UINT8)
    if pa.types.is_uint16(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.UINT16)
    if pa.types.is_uint32(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.UINT32)
    if pa.types.is_uint64(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.UINT64)
    if pa.types.is_int8(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.INT8)
    if pa.types.is_int16(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.INT16)
    if pa.types.is_int32(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.INT32)
    if pa.types.is_int64(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.INT64)
    if pa.types.is_float16(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.FLOAT16)
    if pa.types.is_float32(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.FLOAT32)
    if pa.types.is_float64(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.FLOAT64)
    if pa.types.is_time32(pa_dtype):
        return _TimeUnitDType(type_code=_DTypeCode.TIME32, time_unit=_TimeUnitCode(pa_dtype.unit))

    if pa.types.is_time64(pa_dtype):
        return _TimeUnitDType(type_code=_DTypeCode.TIME64, time_unit=_TimeUnitCode(pa_dtype.unit))

    if pa.types.is_timestamp(pa_dtype):
        return _TimestampDType(
            type_code=_DTypeCode.TIMESTAMP,
            time_unit=_TimeUnitCode(pa_dtype.unit),
            timezone=pa_dtype.tz,
        )

    if pa.types.is_date32(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.DATE32)
    if pa.types.is_date64(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.DATE64)
    if pa.types.is_duration(pa_dtype):
        return _TimeUnitDType(type_code=_DTypeCode.DURATION, time_unit=_TimeUnitCode(pa_dtype.unit))

    if pa.types.is_large_string(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.LARGE_STRING)
    if pa.types.is_large_binary(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.LARGE_BINARY)
    if pa.types.is_fixed_size_binary(pa_dtype):
        return _BinaryDType(type_code=_DTypeCode.BINARY, length=pa_dtype.byte_width)
    if pa.types.is_binary(pa_dtype):
        return _BinaryDType(type_code=_DTypeCode.BINARY, length=-1)
    if pa.types.is_string(pa_dtype):
        return _SingletonDType(type_code=_DTypeCode.STRING)

    if pa.types.is_large_list(pa_dtype):
        return _LargeListDType(
            type_code=_DTypeCode.LARGE_LIST, inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type)
        )
    if pa.types.is_fixed_size_list(pa_dtype):
        return _ListDType(
            type_code=_DTypeCode.LIST,
            inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type),
            length=pa_dtype.list_size,
        )

    if pa.types.is_list(pa_dtype):
        return _ListDType(
            type_code=_DTypeCode.LIST,
            inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type),
            length=-1,
        )

    if pa.types.is_struct(pa_dtype):
        schema = pa.schema(pa_dtype)
        return _StructDType(
            type_code=_DTypeCode.STRUCT,
            fields=[
                _StructField(name=name, dtype=_serialize_pyarrow_dtype(dtype))
                for (name, dtype) in zip(schema.names, schema.types)
            ],
        )

    raise ValueError(f"Unsupported dtype: {pa_dtype}")



from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel

from chalk.integrations.named import load_integration_variable
from chalk.sink._models import SinkIntegrationProtocol
from chalk.streams.base import StreamSource
from chalk.utils.duration import Duration
from chalk.utils.string import comma_whitespace_split


class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True):
    bootstrap_server: Optional[Union[str, List[str]]] = None
    """The URL of one of your Kafka brokers from which to fetch initial metadata about your Kafka cluster"""

    topic: Optional[Union[str, List[str]]] = None
    """The name of the topic to subscribe to."""

    ssl_keystore_location: Optional[str] = None
    """
    An S3 or GCS URI that points to the keystore file that should be
    used for brokers. You must configure the appropriate AWS or
    GCP integration in order for Chalk to be able to access these
    files.
    """

    ssl_ca_file: Optional[str] = None,
    """
    An S3 or GCS URI that points to the certificate authority file that should be
    used to verify broker certificates. You must configure the appropriate AWS or
    GCP integration in order for Chalk to be able to access these files.
    """


    client_id_prefix: str = "chalk/"
    group_id_prefix: str = "chalk/"

    security_protocol: Literal["PLAINTEXT", "SSL", "SASL_PLAINTEXT", "SASL_SSL"] = "PLAINTEXT"
    """
    Protocol used to communicate with brokers.
    Valid values are `"PLAINTEXT"`, `"SSL"`, `"SASL_PLAINTEXT"`, and `"SASL_SSL"`.
    Defaults to `"PLAINTEXT"`.
    """

    sasl_mechanism: Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"] = "PLAIN"
    """
    Authentication mechanism when `security_protocol`
    is configured for SASL_PLAINTEXT or SASL_SSL.
    Valid values are `"PLAIN"`, `"GSSAPI"`, `"SCRAM-SHA-256"`, `"SCRAM-SHA-512"`, `"OAUTHBEARER"`.
    Defaults to `"PLAIN"`.
    """

    sasl_username: Optional[str] = None
    """
    Username for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
    """

    sasl_password: Optional[str] = None
    """
    Password for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
    """

    name: Optional[str] = None
    """
    The name of the integration, as configured in your Chalk Dashboard.
    """

    late_arrival_deadline: Duration = "infinity"
    """
    Messages older than this deadline will not be processed.
    """

    def __init__(
        self,
        *,
        bootstrap_server: Optional[Union[str, List[str]]] = None,
        topic: Optional[Union[str, List[str]]] = None,
        ssl_keystore_location: Optional[str] = None,
        ssl_ca_file: Optional[str] = None,
        client_id_prefix: Optional[str] = None,
        group_id_prefix: Optional[str] = None,
        security_protocol: Optional[str] = None,
        sasl_mechanism: Optional[Literal["PLAIN", "GSSAPI", "SCRAM-SHA-256", "SCRAM-SHA-512", "OAUTHBEARER"]] = None,
        sasl_username: Optional[str] = None,
        sasl_password: Optional[str] = None,
        name: Optional[str] = None,
        late_arrival_deadline: Duration = "infinity",
    ):
        super(KafkaSource, self).__init__(
            bootstrap_server=bootstrap_server
            or load_integration_variable(
                name="KAFKA_BOOTSTRAP_SERVER", integration_name=name, parser=comma_whitespace_split
            ),
            topic=topic
            or load_integration_variable(name="KAFKA_TOPIC", integration_name=name, parser=comma_whitespace_split),
            ssl_keystore_location=ssl_keystore_location
            or load_integration_variable(name="KAFKA_SSL_KEYSTORE_LOCATION", integration_name=name),
            client_id_prefix=client_id_prefix
            or load_integration_variable(name="KAFKA_CLIENT_ID_PREFIX", integration_name=name)
            or KafkaSource.__fields__.get("client_id_prefix").default,
            group_id_prefix=group_id_prefix
            or load_integration_variable(name="KAFKA_GROUP_ID_PREFIX", integration_name=name)
            or KafkaSource.__fields__.get("group_id_prefix").default,
            security_protocol=security_protocol
            or load_integration_variable(name="KAFKA_SECURITY_PROTOCOL", integration_name=name)
            or KafkaSource.__fields__.get("security_protocol").default,
            sasl_mechanism=sasl_mechanism
            or load_integration_variable(name="KAFKA_SASL_MECHANISM", integration_name=name)
            or KafkaSource.__fields__.get("sasl_mechanism").default,
            sasl_username=sasl_username or load_integration_variable(name="KAFKA_SASL_USERNAME", integration_name=name),
            sasl_password=sasl_password or load_integration_variable(name="KAFKA_SASL_PASSWORD", integration_name=name),
            name=name,
            late_arrival_deadline=late_arrival_deadline,
            ssl_ca_file=ssl_ca_file or load_integration_variable(name="KAFKA_SSL_CA_FILE", integration_name=name),
        )

    def _config_to_json(self) -> Any:
        return self.json()
